########################################################################################
#
#    Author: Miika Nissi
#    Copyright 2023-2023 Miika Nissi (https://miikanissi.com)
#
#    This file is part of zebrafy
#    (see https://github.com/miikanissi/zebrafy).
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public License
#    along with this program. If not, see <http://www.gnu.org/licenses/>.
#
########################################################################################

# 1. Standard library imports:
import base64
import io
import operator
import re
import sys
import zlib

# 2. Known third party imports:
from PIL import Image
from pypdfium2 import PdfDocument, PdfImage, PdfMatrix

# 3. Local imports in the relative form:
from .crc import CRC

GF_MATCHER = re.compile(
    r"\^GF([ABC]*),([1-9][0-9]*),([1-9][0-9]*),([1-9][0-9]*),(.*?(?=\^FS))\^FS"
)

if sys.version_info >= (3, 9):
    DimensionsType = tuple[int, int]
    ToImagesType = list[Image.Image]
else:
    from typing import List, Tuple

    ToImagesType = List[Image.Image]
    DimensionsType = Tuple[int, int]


class ZebrafyZPL:
    """
    Provides a method for converting Zebra Programming Language (ZPL) graphic fields \
    to PDF and images.

    :param zpl_data: A valid ZPL string.
    """

    def __init__(self, zpl_data: str):
        self.zpl_data = zpl_data

    zpl_data = property(operator.attrgetter("_zpl_data"))

    @zpl_data.setter
    def zpl_data(self, d):
        if d is None:
            raise ValueError("ZPL data cannot be empty.")
        if not isinstance(d, str):
            raise TypeError(
                "ZPL data must be a valid ZPL string. {param_type} was given.".format(
                    param_type=type(d)
                )
            )
        self._zpl_data = d

    def _match_dimensions(self, total: int, width: int) -> DimensionsType:
        """
        Get image dimensions from ZPL graphic field.

        :param total: Total number of bytes comprising the graphic format.
        :param width: Total number of bytes comprising one row of the data.
        :returns: A tuple of integers containing the width and height of the \
        graphic field image.
        """

        return int(width * 8), int(total / width)

    def to_images(self) -> ToImagesType:
        """
        Converts Zebra Programming Language (ZPL) graphic fields to PIL Image objects.

        :returns: A list containing PIL Images converted from ZPL graphic fields.
        """
        matches = GF_MATCHER.findall(self._zpl_data)
        if not matches:
            raise ValueError("Could not find a graphic field (^GF) in ZPL content.")

        pil_images = []
        for match in matches:
            width, height = self._match_dimensions(int(match[2]), int(match[3]))
            compression_type = match[0].upper()
            data_bytes = match[4]

            # Compression types B and C have base64 encoded data
            if (compression_type == "C" and data_bytes.startswith(":Z64")) or (
                compression_type == "B" and data_bytes.startswith(":B64")
            ):
                crc = data_bytes[-4:]
                data_bytes = data_bytes[5:-5]

                # Validate CRC to ensure data bytes are valid and unchanged
                if crc != CRC(data_bytes.encode("ascii")).get_crc_hex_string():
                    raise ValueError("CRC mismatch.")

                data_bytes = base64.b64decode(data_bytes)

                # Compression type C: decompress LZ77 / Zlib compression
                if compression_type == "C":
                    data_bytes = zlib.decompress(data_bytes)

            # Compression type A contains ASCII hexadecimal data
            elif compression_type == "A":
                data_bytes = bytes.fromhex(data_bytes)

            else:
                raise ValueError(
                    "No valid compression type found in ZPL graphic field (^GF)."
                )

            pil_image = Image.frombytes("1", (width, height), data_bytes)
            pil_images.append(pil_image)

        return pil_images

    def to_pdf(self) -> bytes:
        """
        Converts Zebra Programming Language (ZPL) graphic fields to PDF.

        :returns: PDF bytes from ZPL graphic fields.
        """
        pil_images = self.to_images()
        pdf = PdfDocument.new()

        for pil_image in pil_images:
            # Save PIL Image as JPEG bytes for pypdfium2 to convert into PDF.
            image_bytes = io.BytesIO()
            pil_image.save(image_bytes, format="JPEG")

            # Load image bytes into PDF Image using pypdfium2 and get size
            image = PdfImage.new(pdf)
            image.load_jpeg(image_bytes)
            width, height = image.get_size()
            matrix = PdfMatrix().scale(width, height)
            image.set_matrix(matrix)

            # Save PDF Image on a new page on the PDF.
            page = pdf.new_page(width, height)
            page.insert_obj(image)
            page.gen_content()

        pdf_bytes = io.BytesIO()
        # pypdfium2.PdfDocument save method not to be confused with PIL.Image save method
        pdf.save(pdf_bytes)

        return pdf_bytes.getvalue()



########################################################################################
#
#    Author: Miika Nissi
#    Copyright 2023-2023 Miika Nissi (https://miikanissi.com)
#
#    This file is part of zebrafy
#    (see https://github.com/miikanissi/zebrafy).
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public License
#    along with this program. If not, see <http://www.gnu.org/licenses/>.
#
########################################################################################

# 1. Standard library imports:
import operator

# 2. Known third party imports:

# 3. Local imports in the relative form:


class CRC:
    """
    Utility class to calculate CRC-16-CCITT algorithm across the received data bytes.

    CRC-16-CCITT polynomial representation: x^{16} + x^{12} + x^5 + 1

    :param data_bytes: Bytes object for which to calculate CRC
    :param poly: Reversed polynomial representation for CRC-16-CCITT calculation, \
    defaults to ``0x8408``
    """

    def __init__(self, data_bytes: bytes, poly: int = None):
        self.data_bytes = data_bytes
        if poly is None:
            poly = 0x8408
        self.poly = poly

    data_bytes = property(operator.attrgetter("_data_bytes"))

    @data_bytes.setter
    def data_bytes(self, d):
        if d is None:
            raise ValueError("Bytes data cannot be empty.")
        if not isinstance(d, bytes):
            raise TypeError(
                "Bytes data must be a valid bytes object. {param_type} was given."
                .format(param_type=type(d))
            )
        self._data_bytes = d

    poly = property(operator.attrgetter("_poly"))

    @poly.setter
    def poly(self, p):
        if p is None:
            raise ValueError("Polynomial cannot be empty.")
        if not isinstance(p, int):
            raise TypeError(
                "Polynomial must be a valid integer. {param_type} was given.".format(
                    param_type=type(p)
                )
            )
        self._poly = p

    def _get_crc16_ccitt(self) -> int:
        """
        Calculate CRC-16-CCITT Algorithm.

        :returns: CRC-16-CCITT
        """
        data = bytearray(self._data_bytes)
        crc = 0xFFFF
        for b in data:
            cur_byte = 0xFF & b
            for _ in range(0, 8):
                if (crc & 0x0001) ^ (cur_byte & 0x0001):
                    crc = (crc >> 1) ^ self._poly
                else:
                    crc >>= 1
                cur_byte >>= 1
        crc = ~crc & 0xFFFF
        crc = (crc << 8) | ((crc >> 8) & 0xFF)

        return crc & 0xFFFF

    def get_crc_hex_string(self) -> str:
        """
        Get CRC-16-CCITT as four digit zero padding hexadecimal string.

        :returns: CRC-16-CCITT as four digit zero padding hexadecimal string
        """
        return "%04X" % self._get_crc16_ccitt()



########################################################################################
#
#    Author: Miika Nissi
#    Copyright 2023-2023 Miika Nissi (https://miikanissi.com)
#
#    This file is part of zebrafy
#    (see https://github.com/miikanissi/zebrafy).
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public License
#    along with this program. If not, see <http://www.gnu.org/licenses/>.
#
########################################################################################

# 1. Standard library imports:
import base64
import operator
import zlib

# 2. Known third party imports:
from PIL.Image import Image

# 3. Local imports in the relative form:
from .crc import CRC


class GraphicField:
    """
    Converts a PIL image to Zebra Programming Language (ZPL) graphic field data.

    :param PIL.Image.Image image: An instance of a PIL Image.
    :param compression_type: ZPL compression type parameter that accepts the \
    following values, defaults to ``"A"``:

        - ``"A"``: ASCII hexadecimal - most compatible (default)
        - ``"B"``: Base64 binary
        - ``"C"``: LZ77 / Zlib compressed base64 binary - best compression
    """

    def __init__(self, pil_image: Image, compression_type: str = None):
        self.pil_image = pil_image
        if compression_type is None:
            compression_type = "a"
        self.compression_type = compression_type.upper()

    pil_image = property(operator.attrgetter("_pil_image"))

    @pil_image.setter
    def pil_image(self, i):
        if not i:
            raise ValueError("Image cannot be empty.")
        if not isinstance(i, Image):
            raise TypeError(
                "Image must be a valid PIL.Image.Image object. {param_type} was given."
                .format(param_type=type(i))
            )
        self._pil_image = i

    compression_type = property(operator.attrgetter("_compression_type"))

    @compression_type.setter
    def compression_type(self, c):
        if c is None:
            raise ValueError("Compression type cannot be empty.")
        if not isinstance(c, str):
            raise TypeError(
                "Compression type must be a valid string. {param_type} was given."
                .format(param_type=type(c))
            )
        if c not in ["A", "B", "C"]:
            raise ValueError(
                'Compression type must be "A","B", or "C". {param} was given.'.format(
                    param=c
                )
            )
        self._compression_type = c

    def _get_binary_byte_count(self) -> int:
        """
        Get binary byte count.

        This is the total number of bytes to be transmitted for the total image or
        the total number of bytes that follow parameter bytes_per_row. For ASCII \
        download, the parameter should match parameter graphic_field_count. \
        Out-of-range values are set to the nearest limit.

        :returns: Binary byte count
        """
        return len(self._get_data_string())

    def _get_bytes_per_row(self) -> int:
        """
        Get bytes per row.

        This is the number of bytes in the image data that comprise one row of the \
        image.

        :returns: Bytes per row
        """
        return int((self._pil_image.size[0] + 7) / 8)

    def _get_graphic_field_count(self) -> int:
        """
        Get graphic field count.

        This is the total number of bytes comprising the image data (width x height).

        :returns: Graphic field count."
        """
        return int(self._get_bytes_per_row() * self._pil_image.size[1])

    def _get_data_string(self) -> str:
        """
        Get graphic field data string depending on compression type.

        :returns: Graphic field data string depending on compression type.
        """
        image_bytes = self._pil_image.tobytes()
        data_string = ""

        # Compression type A: Convert bytes to ASCII hexadecimal
        if self._compression_type == "A":
            data_string = image_bytes.hex()

        # Compression type B: Convert bytes to base64 and add header + CRC
        elif self._compression_type == "B":
            b64_bytes = base64.b64encode(image_bytes)
            data_string = ":B64:{encoded_data}:{crc}".format(
                encoded_data=b64_bytes.decode("ascii"),
                crc=CRC(b64_bytes).get_crc_hex_string(),
            )

        # Compression type C: Convert LZ77/ Zlib compressed bytes to base64 and add
        # header + CRC
        elif self._compression_type == "C":
            z64_bytes = base64.b64encode(zlib.compress(image_bytes))
            data_string = ":Z64:{encoded_data}:{crc}".format(
                encoded_data=z64_bytes.decode("ascii"),
                crc=CRC(z64_bytes).get_crc_hex_string(),
            )

        return data_string

    def get_graphic_field(self) -> str:
        """
        Get a complete graphic field string for ZPL.

        :returns: Complete graphic field string for ZPL.
        """
        return "^GF{comp_type},{bb_count},{gf_count},{bpr},{data}^FS".format(
            comp_type=self._compression_type,
            bb_count=self._get_binary_byte_count(),
            gf_count=self._get_graphic_field_count(),
            bpr=self._get_bytes_per_row(),
            data=self._get_data_string(),
        )



########################################################################################
#
#    Author: Miika Nissi
#    Copyright 2023-2023 Miika Nissi (https://miikanissi.com)
#
#    This file is part of zebrafy
#    (see https://github.com/miikanissi/zebrafy).
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public License
#    along with this program. If not, see <http://www.gnu.org/licenses/>.
#
########################################################################################

# 1. Standard library imports:
import io
import os
import unittest
from typing import Union

# 2. Known third party imports:
from PIL import Image

# 3. Local imports in the relative form:
from zebrafy import CRC, GraphicField, ZebrafyImage, ZebrafyPDF, ZebrafyZPL, __version__


class TestZebrafy(unittest.TestCase):
    def _read_static_file(self, file_name: str) -> Union[bytes, str]:
        """
        Helper method to read a test file from static directory.

        :param file_name: File name of a file in tests/static directory.
        :returns: File contents as bytes or string
        """
        open_mode = "r" if file_name.endswith(".zpl") else "rb"
        with open(
            os.path.join(os.path.join(os.path.dirname(__file__), "static"), file_name),
            open_mode,
        ) as file:
            return file.read()

    def test_version(self):
        """Test package version."""
        self.assertEqual(__version__, "1.0.0")

    ###########
    # CRC Tests
    ###########
    def test_crc_data_bytes(self):
        with self.assertRaises(ValueError):
            CRC(None)
        with self.assertRaises(TypeError):
            CRC(123)

    def test_crc_poly(self):
        crc = CRC(b"Python is fun")
        with self.assertRaises(ValueError):
            crc.poly = None
        with self.assertRaises(TypeError):
            crc.poly = "Test"

    ####################
    # GraphicField Tests
    ####################
    def test_graphic_field_image(self):
        with self.assertRaises(ValueError):
            GraphicField(None)
        with self.assertRaises(TypeError):
            GraphicField(123)

    def test_graphic_field_compression_type(self):
        im = Image.new(mode="RGB", size=(200, 200))
        gf = GraphicField(im)
        with self.assertRaises(ValueError):
            gf.compression_type = None
        with self.assertRaises(TypeError):
            gf.compression_type = 123
        with self.assertRaises(ValueError):
            gf.compression_type = "D"

    ####################
    # ZebrafyImage Tests
    ####################
    # Input validation
    def test_zebrafy_image_image(self):
        """Test ZebrafyImage image input"""
        with self.assertRaises(ValueError):
            ZebrafyImage(None)
        with self.assertRaises(TypeError):
            ZebrafyImage(123)

    def test_zebrafy_image_compression_type(self):
        """Test ZebrafyImage compression_type input"""
        im = Image.new(mode="RGB", size=(200, 200))
        zebrafy_image = ZebrafyImage(im)
        with self.assertRaises(ValueError):
            zebrafy_image.compression_type = None
        with self.assertRaises(TypeError):
            zebrafy_image.compression_type = 123
        with self.assertRaises(ValueError):
            zebrafy_image.compression_type = "D"

    def test_zebrafy_image_invert(self):
        """Test ZebrafyImage invert input"""
        im = Image.new(mode="RGB", size=(200, 200))
        zebrafy_image = ZebrafyImage(im)
        with self.assertRaises(ValueError):
            zebrafy_image.invert = None
        with self.assertRaises(TypeError):
            zebrafy_image.invert = "123"

    def test_zebrafy_image_dither(self):
        """Test ZebrafyImage dither input"""
        im = Image.new(mode="RGB", size=(200, 200))
        zebrafy_image = ZebrafyImage(im)
        with self.assertRaises(ValueError):
            zebrafy_image.dither = None
        with self.assertRaises(TypeError):
            zebrafy_image.dither = "123"

    def test_zebrafy_image_threshold(self):
        """Test ZebrafyImage threshold input"""
        im = Image.new(mode="RGB", size=(200, 200))
        zebrafy_image = ZebrafyImage(im)
        with self.assertRaises(ValueError):
            zebrafy_image.threshold = None
        with self.assertRaises(TypeError):
            zebrafy_image.threshold = "123"
        with self.assertRaises(ValueError):
            zebrafy_image.threshold = -1
        with self.assertRaises(ValueError):
            zebrafy_image.threshold = 256

    def test_zebrafy_image_width(self):
        """Test ZebrafyImage width input"""
        im = Image.new(mode="RGB", size=(200, 200))
        zebrafy_image = ZebrafyImage(im)
        with self.assertRaises(ValueError):
            zebrafy_image.width = None
        with self.assertRaises(TypeError):
            zebrafy_image.width = "123"

    def test_zebrafy_image_height(self):
        """Test ZebrafyImage height input"""
        im = Image.new(mode="RGB", size=(200, 200))
        zebrafy_image = ZebrafyImage(im)
        with self.assertRaises(ValueError):
            zebrafy_image.height = None
        with self.assertRaises(TypeError):
            zebrafy_image.height = "123"

    def test_zebrafy_image_pos_x(self):
        """Test ZebrafyImage pos_x input"""
        im = Image.new(mode="RGB", size=(200, 200))
        zebrafy_image = ZebrafyImage(im)
        with self.assertRaises(ValueError):
            zebrafy_image.pos_x = None
        with self.assertRaises(TypeError):
            zebrafy_image.pos_x = "123"

    def test_zebrafy_image_pos_y(self):
        """Test ZebrafyImage pos_y input"""
        im = Image.new(mode="RGB", size=(200, 200))
        zebrafy_image = ZebrafyImage(im)
        with self.assertRaises(ValueError):
            zebrafy_image.pos_y = None
        with self.assertRaises(TypeError):
            zebrafy_image.pos_y = "123"

    def test_zebrafy_image_complete_zpl(self):
        """Test ZebrafyImage complete_zpl input"""
        im = Image.new(mode="RGB", size=(200, 200))
        zebrafy_image = ZebrafyImage(im)
        with self.assertRaises(ValueError):
            zebrafy_image.complete_zpl = None
        with self.assertRaises(TypeError):
            zebrafy_image.complete_zpl = "123"

    # Output validation
    def test_image_to_default_zpl(self):
        """Test image to ZPL with default options."""
        default_zpl = ZebrafyImage(self._read_static_file("test_image.png")).to_zpl()
        self.assertEqual(default_zpl, self._read_static_file("test_image_gfa.zpl"))

    def test_image_to_gfa_zpl(self):
        """Test image to ZPL with A (ASCII) compression."""
        gfa_zpl = ZebrafyImage(
            self._read_static_file("test_image.png"), compression_type="A"
        ).to_zpl()
        self.assertEqual(gfa_zpl, self._read_static_file("test_image_gfa.zpl"))

    def test_image_to_gfb_zpl(self):
        """Test image to ZPL with B (B64 Binary) compression."""
        gfb_zpl = ZebrafyImage(
            self._read_static_file("test_image.png"), compression_type="B"
        ).to_zpl()
        self.assertEqual(gfb_zpl, self._read_static_file("test_image_gfb.zpl"))

    def test_image_to_gfc_zpl(self):
        """Test image to ZPL with C (Z64 Binary) compression."""
        gfc_zpl = ZebrafyImage(
            self._read_static_file("test_image.png"), compression_type="C"
        ).to_zpl()
        self.assertEqual(gfc_zpl, self._read_static_file("test_image_gfc.zpl"))

    def test_image_to_zpl_invert(self):
        """Test image to ZPL inverting the image."""
        gf_zpl = ZebrafyImage(
            self._read_static_file("test_image.png"), invert=True
        ).to_zpl()
        self.assertEqual(gf_zpl, self._read_static_file("test_image_invert.zpl"))

    def test_image_to_zpl_invert_no_dither(self):
        """Test image to ZPL without dithering and inverting the image."""
        gf_zpl = ZebrafyImage(
            self._read_static_file("test_image.png"), dither=False, invert=True
        ).to_zpl()
        self.assertEqual(
            gf_zpl, self._read_static_file("test_image_invert_no_dither.zpl")
        )

    def test_image_to_zpl_no_dither(self):
        """Test image to ZPL without dithering the image."""
        gf_zpl = ZebrafyImage(
            self._read_static_file("test_image.png"), dither=False
        ).to_zpl()
        self.assertEqual(gf_zpl, self._read_static_file("test_image_no_dither.zpl"))

    def test_image_to_zpl_threshold_low(self):
        """Test image to ZPL without dithering the image and low threshold."""
        gf_zpl = ZebrafyImage(
            self._read_static_file("test_image.png"), dither=False, threshold=40
        ).to_zpl()
        self.assertEqual(gf_zpl, self._read_static_file("test_image_low_threshold.zpl"))

    def test_image_to_zpl_threshold_high(self):
        """Test image to ZPL without dithering the image and high threshold."""
        gf_zpl = ZebrafyImage(
            self._read_static_file("test_image.png"), dither=False, threshold=215
        ).to_zpl()
        self.assertEqual(
            gf_zpl, self._read_static_file("test_image_high_threshold.zpl")
        )

    def test_image_to_zpl_width_height(self):
        """Test image to ZPL with width and height."""
        gf_zpl = ZebrafyImage(
            self._read_static_file("test_image.png"), width=500, height=500
        ).to_zpl()
        self.assertEqual(gf_zpl, self._read_static_file("test_image_width_height.zpl"))

    def test_image_to_zpl_pos_x_pos_y(self):
        """Test image to ZPL with pos_x and pos_y."""
        gf_zpl = ZebrafyImage(
            self._read_static_file("test_image.png"), pos_x=100, pos_y=200
        ).to_zpl()
        self.assertEqual(gf_zpl, self._read_static_file("test_image_pos_x_pos_y.zpl"))

    def test_multiple_image_to_zpl(self):
        """Test multiple images to ZPL with default options."""
        gf_zpl = ZebrafyImage(
            self._read_static_file("test_image.png"),
            complete_zpl=False,
        ).to_zpl()
        complete_zpl = "^XA\n" + gf_zpl + "\n" + gf_zpl + "\n^XZ\n"
        self.assertEqual(
            complete_zpl, self._read_static_file("test_image_multiple.zpl")
        )

    ##################
    # ZebrafyPDF Tests
    ##################
    # Input validation
    def test_zebrafy_pdf_pdf_bytes(self):
        """Test ZebrafyImage pdf_bytes input"""
        with self.assertRaises(ValueError):
            ZebrafyPDF(None)
        with self.assertRaises(TypeError):
            ZebrafyPDF(123)

    def test_zebrafy_pdf_compression_type(self):
        """Test ZebrafyPDF compression_type input"""
        pdf = self._read_static_file("test_pdf.pdf")
        zebrafy_pdf = ZebrafyPDF(pdf)
        with self.assertRaises(ValueError):
            zebrafy_pdf.compression_type = None
        with self.assertRaises(TypeError):
            zebrafy_pdf.compression_type = 123
        with self.assertRaises(ValueError):
            zebrafy_pdf.compression_type = "D"

    def test_zebrafy_pdf_invert(self):
        """Test ZebrafyPDF invert input"""
        pdf = self._read_static_file("test_pdf.pdf")
        zebrafy_pdf = ZebrafyPDF(pdf)
        with self.assertRaises(ValueError):
            zebrafy_pdf.invert = None
        with self.assertRaises(TypeError):
            zebrafy_pdf.invert = "123"

    def test_zebrafy_pdf_dither(self):
        """Test ZebrafyPDF dither input"""
        pdf = self._read_static_file("test_pdf.pdf")
        zebrafy_pdf = ZebrafyPDF(pdf)
        with self.assertRaises(ValueError):
            zebrafy_pdf.dither = None
        with self.assertRaises(TypeError):
            zebrafy_pdf.dither = "123"

    def test_zebrafy_pdf_threshold(self):
        """Test ZebrafyPDF threshold input"""
        pdf = self._read_static_file("test_pdf.pdf")
        zebrafy_pdf = ZebrafyPDF(pdf)
        with self.assertRaises(ValueError):
            zebrafy_pdf.threshold = None
        with self.assertRaises(TypeError):
            zebrafy_pdf.threshold = "123"
        with self.assertRaises(ValueError):
            zebrafy_pdf.threshold = -1
        with self.assertRaises(ValueError):
            zebrafy_pdf.threshold = 256

    def test_zebrafy_pdf_width(self):
        """Test ZebrafyPDF width input"""
        pdf = self._read_static_file("test_pdf.pdf")
        zebrafy_pdf = ZebrafyPDF(pdf)
        with self.assertRaises(ValueError):
            zebrafy_pdf.width = None
        with self.assertRaises(TypeError):
            zebrafy_pdf.width = "123"

    def test_zebrafy_pdf_height(self):
        """Test ZebrafyPDF height input"""
        pdf = self._read_static_file("test_pdf.pdf")
        zebrafy_pdf = ZebrafyPDF(pdf)
        with self.assertRaises(ValueError):
            zebrafy_pdf.height = None
        with self.assertRaises(TypeError):
            zebrafy_pdf.height = "123"

    def test_zebrafy_pdf_pos_x(self):
        """Test ZebrafyPDF pos_x input"""
        pdf = self._read_static_file("test_pdf.pdf")
        zebrafy_pdf = ZebrafyPDF(pdf)
        with self.assertRaises(ValueError):
            zebrafy_pdf.pos_x = None
        with self.assertRaises(TypeError):
            zebrafy_pdf.pos_x = "123"

    def test_zebrafy_pdf_pos_y(self):
        """Test ZebrafyPDF pos_y input"""
        pdf = self._read_static_file("test_pdf.pdf")
        zebrafy_pdf = ZebrafyPDF(pdf)
        with self.assertRaises(ValueError):
            zebrafy_pdf.pos_y = None
        with self.assertRaises(TypeError):
            zebrafy_pdf.pos_y = "123"

    def test_zebrafy_pdf_complete_zpl(self):
        """Test ZebrafyPDF complete_zpl input"""
        pdf = self._read_static_file("test_pdf.pdf")
        zebrafy_pdf = ZebrafyPDF(pdf)
        with self.assertRaises(ValueError):
            zebrafy_pdf.complete_zpl = None
        with self.assertRaises(TypeError):
            zebrafy_pdf.complete_zpl = "123"

    def test_pdf_to_default_zpl(self):
        """Test PDF to ZPL with default options."""
        default_zpl = ZebrafyPDF(self._read_static_file("test_pdf.pdf")).to_zpl()
        self.assertEqual(default_zpl, self._read_static_file("test_pdf_gfa.zpl"))

    def test_pdf_to_gfa_zpl(self):
        """Test PDF to ZPL with A (ASCII) compression."""
        gfa_zpl = ZebrafyPDF(
            self._read_static_file("test_pdf.pdf"), compression_type="A"
        ).to_zpl()
        self.assertEqual(gfa_zpl, self._read_static_file("test_pdf_gfa.zpl"))

    def test_pdf_to_gfb_zpl(self):
        """Test PDF to ZPL with B (B64 Binary) compression."""
        gfb_zpl = ZebrafyPDF(
            self._read_static_file("test_pdf.pdf"), compression_type="B"
        ).to_zpl()
        self.assertEqual(gfb_zpl, self._read_static_file("test_pdf_gfb.zpl"))

    def test_pdf_to_gfc_zpl(self):
        """Test PDF to ZPL with C (Z64 Binary) compression."""
        gfc_zpl = ZebrafyPDF(
            self._read_static_file("test_pdf.pdf"), compression_type="C"
        ).to_zpl()
        self.assertEqual(gfc_zpl, self._read_static_file("test_pdf_gfc.zpl"))

    def test_pdf_to_zpl_no_dither(self):
        """Test PDF to ZPL without dithering the PDF."""
        gf_zpl = ZebrafyPDF(
            self._read_static_file("test_pdf.pdf"), dither=False
        ).to_zpl()
        self.assertEqual(gf_zpl, self._read_static_file("test_pdf_no_dither.zpl"))

    def test_pdf_to_zpl_threshold_low(self):
        """Test PDF to ZPL without dithering the PDF and low threshold."""
        gf_zpl = ZebrafyPDF(
            self._read_static_file("test_pdf.pdf"), dither=False, threshold=40
        ).to_zpl()
        self.assertEqual(gf_zpl, self._read_static_file("test_pdf_low_threshold.zpl"))

    def test_pdf_to_zpl_threshold_high(self):
        """Test PDF to ZPL without dithering the PDF and high threshold."""
        gf_zpl = ZebrafyPDF(
            self._read_static_file("test_pdf.pdf"), dither=False, threshold=215
        ).to_zpl()
        self.assertEqual(gf_zpl, self._read_static_file("test_pdf_high_threshold.zpl"))

    def test_pdf_to_zpl_width_height(self):
        """Test PDF to ZPL without dithering the PDF and high threshold."""
        gf_zpl = ZebrafyPDF(
            self._read_static_file("test_pdf.pdf"), width=720, height=1280
        ).to_zpl()
        self.assertEqual(gf_zpl, self._read_static_file("test_pdf_width_height.zpl"))

    ##################
    # ZebrafyZPL Tests
    ##################
    def test_zebrafy_zpl_zpl_data(self):
        with self.assertRaises(ValueError):
            ZebrafyZPL(None)
        with self.assertRaises(TypeError):
            ZebrafyZPL(123)

    def test_gfa_zpl_to_image(self):
        """Test ZPL GFA to image bytes."""
        image = ZebrafyZPL(self._read_static_file("test_image_gfa.zpl")).to_images()[0]
        image_bytes = io.BytesIO()
        image.save(image_bytes, format="PNG")
        self.assertEqual(
            image_bytes.getvalue(), self._read_static_file("test_image_gfa.png")
        )

    def test_gfb_zpl_to_image(self):
        """Test ZPL GFB to image bytes."""
        image = ZebrafyZPL(self._read_static_file("test_image_gfb.zpl")).to_images()[0]
        image_bytes = io.BytesIO()
        image.save(image_bytes, format="PNG")
        self.assertEqual(
            image_bytes.getvalue(), self._read_static_file("test_image_gfb.png")
        )

    def test_gfc_zpl_to_image(self):
        """Test ZPL GFC to image bytes."""
        image = ZebrafyZPL(self._read_static_file("test_image_gfc.zpl")).to_images()[0]
        image_bytes = io.BytesIO()
        image.save(image_bytes, format="PNG")
        self.assertEqual(
            image_bytes.getvalue(), self._read_static_file("test_image_gfc.png")
        )

    def test_broken_zpl_gf_to_image(self):
        """Test broken ZPL to image bytes - resulting in ValueError."""
        zebrafy_broken_zpl = ZebrafyZPL(
            self._read_static_file("test_image_broken.zpl"),
        )
        with self.assertRaises(ValueError):
            zebrafy_broken_zpl.to_images()

    def test_broken_zpl_gfc_crc_to_image(self):
        """Test broken ZPL gfc to image bytes - resulting in ValueError."""
        zebrafy_broken_zpl = ZebrafyZPL(
            self._read_static_file("test_image_gfc_broken_crc.zpl"),
        )
        with self.assertRaises(ValueError):
            zebrafy_broken_zpl.to_images()

    def test_broken_zpl_gfc_compression_to_image(self):
        """Test broken ZPL gfc to image bytes - resulting in ValueError."""
        zebrafy_broken_zpl = ZebrafyZPL(
            self._read_static_file("test_image_gfc_broken_compression.zpl"),
        )
        with self.assertRaises(ValueError):
            zebrafy_broken_zpl.to_images()

    def test_gfa_zpl_to_pdf(self):
        """Test ZPL GFA to PDF bytes."""
        pdf_bytes = ZebrafyZPL(self._read_static_file("test_pdf_gfa.zpl")).to_pdf()
        gfa_zpl = ZebrafyPDF(pdf_bytes, compression_type="A").to_zpl()
        self.assertEqual(gfa_zpl, self._read_static_file("test_pdf_gfa.zpl"))

    def test_gfb_zpl_to_pdf(self):
        """Test ZPL GFB to PDF bytes."""
        pdf_bytes = ZebrafyZPL(self._read_static_file("test_pdf_gfb.zpl")).to_pdf()
        gfb_zpl = ZebrafyPDF(pdf_bytes, compression_type="B").to_zpl()
        self.assertEqual(gfb_zpl, self._read_static_file("test_pdf_gfb.zpl"))

    def test_gfc_zpl_to_pdf(self):
        """Test ZPL GFC to PDF bytes."""
        pdf_bytes = ZebrafyZPL(self._read_static_file("test_pdf_gfc.zpl")).to_pdf()
        gfc_zpl = ZebrafyPDF(pdf_bytes, compression_type="C").to_zpl()
        self.assertEqual(gfc_zpl, self._read_static_file("test_pdf_gfc.zpl"))



########################################################################################
#
#    Author: Miika Nissi
#    Copyright 2023-2023 Miika Nissi (https://miikanissi.com)
#
#    This file is part of zebrafy
#    (see https://github.com/miikanissi/zebrafy).
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public License
#    along with this program. If not, see <http://www.gnu.org/licenses/>.
#
########################################################################################

# 1. Standard library imports:
import operator

# 2. Known third party imports:
from pypdfium2 import PdfDocument

# 3. Local imports in the relative form:
from .zebrafy_image import ZebrafyImage


class ZebrafyPDF:
    """
    Provides a method for converting PDFs to Zebra Programming Language (ZPL).

    :param pdf_bytes: PDF as a bytes object.
    :param compression_type: ZPL compression type parameter that accepts the \
    following values, defaults to ``"A"``:

        - ``"A"``: ASCII hexadecimal - most compatible (default)
        - ``"B"``: Base64 binary
        - ``"C"``: LZ77 / Zlib compressed base64 binary - best compression
    :param invert: Invert the black and white in resulting image, defaults to ``False``
    :param dither: Dither the pixels instead of hard limit on black and white, \
    defaults to ``False``
    :param threshold: Black pixel threshold for undithered PDF (``0-255``), defaults \
    to ``128``
    :param width: Width of the PDF in the resulting ZPL. If ``0``, use default PDF \
    width, defaults to ``0``
    :param height: Height of the PDF in the resulting ZPL. If ``0``, use default \
    PDF height, defaults to ``0``
    :param pos_x: X position of the PDF on the resulting ZPL, defaults to ``0``
    :param pos_y: Y position of the PDF on the resulting ZPL, defaults to ``0``
    :param complete_zpl: Return a complete ZPL with header and footer included. \
    Otherwise return only the graphic field, defaults to ``True``
    """

    def __init__(
        self,
        pdf_bytes: bytes,
        compression_type: str = None,
        invert: bool = None,
        dither: bool = None,
        threshold: int = None,
        width: int = None,
        height: int = None,
        pos_x: int = None,
        pos_y: int = None,
        complete_zpl: bool = None,
    ):
        self.pdf_bytes = pdf_bytes
        if compression_type is None:
            compression_type = "a"
        self.compression_type = compression_type.upper()
        if invert is None:
            invert = False
        self.invert = invert
        if dither is None:
            dither = True
        self.dither = dither
        if threshold is None:
            threshold = 128
        self.threshold = threshold
        if width is None:
            width = 0
        self.width = width
        if height is None:
            height = 0
        self.height = height
        if pos_x is None:
            pos_x = 0
        self.pos_x = pos_x
        if pos_y is None:
            pos_y = 0
        self.pos_y = pos_y
        if complete_zpl is None:
            complete_zpl = True
        self.complete_zpl = complete_zpl

    pdf_bytes = property(operator.attrgetter("_pdf_bytes"))

    @pdf_bytes.setter
    def pdf_bytes(self, p):
        if not p:
            raise ValueError("PDF cannot be empty.")
        if not isinstance(p, bytes):
            raise TypeError(
                "PDF must be a valid bytes object {param_type} was given.".format(
                    param_type=type(p)
                )
            )
        self._pdf_bytes = p

    compression_type = property(operator.attrgetter("_compression_type"))

    @compression_type.setter
    def compression_type(self, c):
        if c is None:
            raise ValueError("Compression type cannot be empty.")
        if not isinstance(c, str):
            raise TypeError(
                "Compression type must be a valid string. {param_type} was given."
                .format(param_type=type(c))
            )
        if c not in ["A", "B", "C"]:
            raise ValueError(
                'Compression type must be "A","B", or "C". {param} was given.'.format(
                    param=c
                )
            )
        self._compression_type = c

    invert = property(operator.attrgetter("_invert"))

    @invert.setter
    def invert(self, i):
        if i is None:
            raise ValueError("Invert cannot be empty.")
        if not isinstance(i, bool):
            raise TypeError(
                "Invert must be a boolean. {param_type} was given.".format(
                    param_type=type(i)
                )
            )
        self._invert = i

    dither = property(operator.attrgetter("_dither"))

    @dither.setter
    def dither(self, d):
        if d is None:
            raise ValueError("Dither cannot be empty.")
        if not isinstance(d, bool):
            raise TypeError(
                "Dither must be a boolean. {param_type} was given.".format(
                    param_type=type(d)
                )
            )
        self._dither = d

    threshold = property(operator.attrgetter("_threshold"))

    @threshold.setter
    def threshold(self, t):
        if t is None:
            raise ValueError("Threshold cannot be empty.")
        if not isinstance(t, int):
            raise TypeError(
                "Threshold must be an integer. {param_type} was given.".format(
                    param_type=type(t)
                )
            )
        if t < 0 or t > 255:
            raise ValueError(
                "Threshold must be within 0 to 255. {param} was given.".format(param=t)
            )
        self._threshold = t

    width = property(operator.attrgetter("_width"))

    @width.setter
    def width(self, w):
        if w is None:
            raise ValueError("Width cannot be empty.")
        if not isinstance(w, int):
            raise TypeError(
                "Width must be an integer. {param_type} was given.".format(
                    param_type=type(w)
                )
            )
        self._width = w

    height = property(operator.attrgetter("_height"))

    @height.setter
    def height(self, h):
        if h is None:
            raise ValueError("Height cannot be empty.")
        if not isinstance(h, int):
            raise TypeError(
                "Height must be an integer. {param_type} was given.".format(
                    param_type=type(h)
                )
            )
        self._height = h

    pos_x = property(operator.attrgetter("_pos_x"))

    @pos_x.setter
    def pos_x(self, x):
        if x is None:
            raise ValueError("X position cannot be empty.")
        if not isinstance(x, int):
            raise TypeError(
                "X position must be an integer. {param_type} was given.".format(
                    param_type=type(x)
                )
            )
        self._pos_x = x

    pos_y = property(operator.attrgetter("_pos_y"))

    @pos_y.setter
    def pos_y(self, y):
        if y is None:
            raise ValueError("Y position cannot be empty.")
        if not isinstance(y, int):
            raise TypeError(
                "Y position must be an integer. {param_type} was given.".format(
                    param_type=type(y)
                )
            )
        self._pos_y = y

    complete_zpl = property(operator.attrgetter("_complete_zpl"))

    @complete_zpl.setter
    def complete_zpl(self, c):
        if c is None:
            raise ValueError("Complete ZPL cannot be empty.")
        if not isinstance(c, bool):
            raise TypeError(
                "Complete ZPL must be a boolean. {param_type} was given.".format(
                    param_type=type(c)
                )
            )
        self._complete_zpl = c

    def to_zpl(self) -> str:
        """
        Converts PDF bytes to Zebra Programming Language (ZPL).

        :returns: A complete ZPL file string which can be sent to a ZPL compatible \
        printer or a ZPL graphic field if complete_zpl is not set.
        """
        # Open and convert image to grayscale
        pdf = PdfDocument(self._pdf_bytes)
        graphic_fields = ""
        for page in pdf:
            bitmap = page.render(scale=1, rotation=0)
            pil_image = bitmap.to_pil()
            zebrafy_image = ZebrafyImage(
                pil_image,
                compression_type=self._compression_type,
                invert=self._invert,
                dither=self._dither,
                threshold=self._threshold,
                width=self._width,
                height=self._height,
                pos_x=self._pos_x,
                pos_y=self._pos_y,
                complete_zpl=False,
            )
            graphic_fields += zebrafy_image.to_zpl() + "\n"

        if self._complete_zpl:
            return "^XA\n" + graphic_fields + "^XZ\n"

        return graphic_fields
